import torch
from torch.utils.data import DataLoader

from ptls.data_load import padded_collate, ZeroDownSampler, DropoutTrxDataset, TrxDataset, LastKTrxDataset
from ptls.data_load import augmentation_chain
from ptls_tests.utils.data_generation import gen_trx_data


def test_padded_collate():
    data = [
        ({'a': torch.LongTensor([1, 2, 3, 4])}, torch.LongTensor([0])),
        ({'a': torch.LongTensor([1, 2])},  torch.LongTensor([0])),
        ({'a': torch.LongTensor([1])},  torch.LongTensor([1])),
    ]

    tt = torch.LongTensor([
        [1, 2, 3, 4],
        [1, 2, 0, 0],
        [1, 0, 0, 0]])

    x, y = padded_collate(data)

    assert x.payload['a'].shape == (3, 4)
    assert x.payload['a'].eq(tt).all()


def test_zero_down_sampler():
    y = torch.LongTensor([1, 0, 1, 0, 0, 0])
    sampler = ZeroDownSampler(y)
    idx = list(sampler)
    assert sum(y[idx]) == 2
    assert len(y[idx]) == 4


def test_data_loader():
    data = gen_trx_data((torch.rand(1000)*60+1).long())
    y0 = torch.LongTensor([e['target'] for e in data])
    ds = DropoutTrxDataset(TrxDataset(data), trx_dropout=0, seq_len=15)
    dl = DataLoader(ds, 10, collate_fn=padded_collate)
    y = torch.cat([y for x, y in dl])
    assert all(y0 == y)


def test_last_k_trx_dataset():
    data = gen_trx_data([100, 100, 100])
    res = [len(next(iter(x.values()))) for x, _ in LastKTrxDataset(TrxDataset(data), .5)]
    assert all(torch.tensor(res) == torch.tensor([50, 50, 50]))
    res = [len(next(iter(x.values()))) for x, _ in LastKTrxDataset(TrxDataset(data), .2)]
    assert all(torch.tensor(res) == torch.tensor([20, 20, 20]))


def _inc(x):
    return x + 1


def _double(x):
    return x * 2


def test_augmentation_chain():
    a = augmentation_chain(_inc, _double, _inc)
    out = a(2)
    assert out == 7


def test_augmentation_chain_pickle():
    import pickle

    a = augmentation_chain(_inc, _double)
    pickle.dumps(a)



import random
import numpy as np

def sequence_pair_augmentation(item, max_lenght=np.inf):
    length = len(next(iter(item.values())))
    l_length = random.randint(length // 4, min(3 * length // 4, max_lenght // 2))
    left = {k: v[:l_length] for k, v in item.items()}
    right = {k: v[l_length:] for k, v in item.items()}
    return left, right  



from copy import deepcopy

import torch

from ptls.data_load.augmentations.random_slice import RandomSlice
from ptls.data_load.utils import collate_feature_dict
from ptls.data_load.feature_dict import FeatureDict
from ptls.data_load.padded_batch import PaddedBatch


class RtdDataset(torch.utils.data.Dataset):
    def __init__(self, data,
                 min_len, max_len, rate_for_min: float = 1.0,
                 replace_prob=0.15,
                 skip_first=1,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)  # required for mixin class

        self.data = data
        self.r_slice = RandomSlice(min_len, max_len, rate_for_min)

        self.replace_prob = replace_prob
        self.skip_first = skip_first

    def __len__(self):
        return len(self.data)

    def __getitem__(self, item):
        feature_arrays = self.data[item]
        return self.process(feature_arrays)

    def __iter__(self):
        for feature_arrays in self.data:
            yield self.process(feature_arrays)

    def process(self, feature_arrays):
        feature_arrays = {k: v for k, v in feature_arrays.items() if FeatureDict.is_seq_feature(k, v)}
        return self.r_slice(feature_arrays)

    def collate_fn(self, batch):
        padded_batch = collate_feature_dict(batch)

        new_x, lengths, mask = padded_batch.payload, padded_batch.seq_lens, padded_batch.seq_len_mask

        to_replace = torch.bernoulli(mask * self.replace_prob).bool()
        to_replace[:, :self.skip_first] = False

        sampled_trx_ids = torch.multinomial(
            mask.flatten().float(),
            num_samples=to_replace.sum().item(),
            replacement=True,
        )

        to_replace_flatten = to_replace.flatten()
        new_x = deepcopy(new_x)
        for k, v in new_x.items():
            if FeatureDict.is_seq_feature(k, v):
                v.flatten()[to_replace_flatten] = v.flatten()[sampled_trx_ids]

        return PaddedBatch(new_x, lengths), to_replace.long().flatten()[mask.flatten().bool()]


class RtdIterableDataset(RtdDataset, torch.utils.data.IterableDataset):
    pass



class PairSelector:
    """
    Implementation should return indices of positive pairs and negative pairs that will be passed to compute
    Contrastive Loss
    return positive_pairs, negative_pairs
    """

    def __init__(self):
        pass

    def get_pairs(self, embeddings, labels):
        raise NotImplementedError


import numpy as np
import torch


class FeatureDict:
    """Tools for feature-dict format
    Feature dict:
        keys are feature names
        values are feature values, sequential, scalar or arrays
    """
    def __init__(self, *args, **kwargs):
        """Mixin constructor
        """
        super().__init__(*args, **kwargs)

    @staticmethod
    def is_seq_feature(k: str, x):
        """Check is value sequential feature
        Synchronized with ptls.data_load.padded_batch.PaddedBatch.is_seq_feature

        Iterables are:
            np.array
            torch.Tensor

        Not iterable:
            list    - dont supports indexing

        Parameters
        ----------
        k:
            feature_name
        x:
            value for check

        Returns
        -------
            True if value is iterable
        """
        if k == 'event_time':
            return True
        if k.startswith('target'):
            return False
        if type(x) in (np.ndarray, torch.Tensor):
            return True
        return False

    @staticmethod
    def seq_indexing(d, ix):
        """Apply indexing for seq_features only

        Parameters
        ----------
        d:
            feature dict
        ix:
            indexes

        Returns
        -------

        """
        return {k: v[ix] if FeatureDict.is_seq_feature(k, v) else v for k, v in d.items()}

    @staticmethod
    def get_seq_len(d):
        """Finds a sequence column and return its length

        Parameters
        ----------
        d:
            feature-dict

        """
        if 'event_time' in d:
            return len(d['event_time'])
        return len(next(v for k, v in d.items() if FeatureDict.is_seq_feature(k, v)))



import pytest

from ptls.data_load.list_splitter import ListSplitter


def test_data_size_select_01():
    data = ListSplitter([f'{i}.csv' for i in range(10)], valid_size=0.5)
    sizes = data.size_select()
    assert sizes == (5, 5)


def test_data_size_select_02():
    data = ListSplitter([f'{i}.csv' for i in range(10)], valid_size=0.05)
    sizes = data.size_select()
    assert sizes == (9, 1)


def test_data_size_select_03():
    data = ListSplitter([f'{i}.csv' for i in range(2)], valid_size=0.05)
    sizes = data.size_select()
    assert sizes == (1, 1)


def test_data_size_select_04():
    with pytest.raises(AttributeError):
        _ = ListSplitter([f'{i}.csv' for i in range(1)], valid_size=0.05)


def test_data_files():
    data = ListSplitter([f'{i}.csv' for i in range(5)], valid_size=0.45, seed=42)
    assert data.valid == ['1.csv', '4.csv']



import random

import numpy as np

from ptls.data_load.feature_dict import FeatureDict


class RandomSlice(FeatureDict):
    """
    This class is used as 'f_augmentation' argument for 
    ptls.data_load.datasets.augmentation_dataset.AugmentationDataset (AugmentationIterableDataset).
    """
    def __init__(self, min_len, max_len, rate_for_min=1.0):
        super().__init__()

        self.min_len = min_len
        self.max_len = max_len
        self.rate_for_min = rate_for_min

    def __call__(self, x):
        seq_len = self.get_seq_len(x)

        idx = self.get_idx(seq_len)
        new_x = self.seq_indexing(x, idx)
        return new_x

    def get_idx(self, seq_len):
        new_idx = np.arange(seq_len)

        min_len, max_len = self.get_min_max(seq_len)
        if max_len < min_len:
            return new_idx
        new_len = random.randint(min_len, max_len)

        avail_pos = seq_len - new_len
        pos = random.randint(0, avail_pos)
        return new_idx[pos:pos+new_len]

    def get_min_max(self, seq_len):
        max_len = int(min(self.max_len, seq_len))
        min_len = int(min(self.min_len, seq_len * self.rate_for_min))
        if min_len < 1:
            min_len = 1
        return min_len, max_len



import pyspark.sql

class ColTransformerPysparkMixin:
    def check_is_col_exists(self, x: pyspark.sql.DataFrame):
        if self.col_name_original not in x.columns:
            raise AttributeError(f'cols_event_time="{self.col_name_original}" not in source dataframe. '
                                 f'Found {x.columns}')

    def drop_original_col(self, x: pyspark.sql.DataFrame):
        if self.col_name_original != self.col_name_target and self.is_drop_original_col:
            x = x.drop(self.col_name_original)
        return x



import warnings

from torch.utils.data import Dataset


class MapAugmentationDataset(Dataset):
    def __init__(self, base_dataset, a_chain=None):
        warnings.warn('Use `ptls.data_load.datasets.AugmentationDataset`', DeprecationWarning)
        self.base_dataset = base_dataset
        self.a_chain = a_chain

    def __len__(self):
        return len(self.base_dataset)

    def __getitem__(self, idx):
        row = self.base_dataset[idx]
        if type(row) is tuple:
            feature_arrays, target = row
            if self.a_chain is not None:
                feature_arrays = self.a_chain(feature_arrays)
            return feature_arrays, target
        else:
            feature_arrays = row
            if self.a_chain is not None:
                feature_arrays = self.a_chain(feature_arrays)
            return feature_arrays



import torch
from omegaconf import OmegaConf

from ptls.nn.trx_encoder import TrxEncoder
from ptls.data_load.padded_batch import PaddedBatch
from ptls.nn.seq_encoder import RnnSeqEncoder, TransformerSeqEncoder
from hydra.utils import instantiate


def get_data():
    payload = {
        'amount': torch.arange(4*10).view(4, 10).float(),
        'event_time': torch.arange(4*10).view(4, 10).float(),
        'mcc_code': torch.arange(4*10).view(4, 10),
        'tr_type': torch.arange(4*10).view(4, 10)
    }
    return PaddedBatch(
        payload=payload,
        length=torch.tensor([4, 2, 6, 8])
    )


def test_rnn_shape():
    model = RnnSeqEncoder(
        trx_encoder=TrxEncoder(
            embeddings={
                'mcc_code': {'in': 200, 'out': 48},
                'tr_type': {'in': 100, 'out': 24},
            },
            numeric_values={'amount': 'identity'},
        ),
        hidden_size=48,
    )

    x = get_data()

    out = model(x)
    assert isinstance(out, torch.Tensor) and out.shape == torch.Size([4, 48])


def test_rnn_hydra_init():
    params = """
seq_encoder: 
    _target_: ptls.nn.RnnSeqEncoder
    trx_encoder:
        _target_: ptls.nn.TrxEncoder
        embeddings:
            mcc_code:
                in: 200
                out: 48
            tr_type:
                in: 100
                out: 24
        numeric_values:
            amount: identity
    hidden_size: 48
    """
    model = instantiate(OmegaConf.create(params))['seq_encoder']

    x = get_data()
    out = model(x)

    assert isinstance(out, torch.Tensor) and out.shape == torch.Size([4, 48])


def test_transformer_shape():
    model = TransformerSeqEncoder(
        trx_encoder=TrxEncoder(
            embeddings={
                'mcc_code': {'in': 200, 'out': 48},
                'tr_type': {'in': 100, 'out': 23},
            },
            numeric_values={'amount': 'identity'},
        ),
        n_heads=2,
        dim_hidden=128,
        n_layers=2,
        max_seq_len=400,
    )

    x = get_data()

    out = model(x)
    assert isinstance(out, torch.Tensor) and out.shape == torch.Size([4, 72])


def test_transformer_hydra_init():
    params = """
seq_encoder: 
    _target_: ptls.nn.TransformerSeqEncoder
    trx_encoder:
        _target_: ptls.nn.TrxEncoder
        embeddings:
            mcc_code:
                in: 200
                out: 48
            tr_type:
                in: 100
                out: 23
        numeric_values:
            amount: identity
    n_heads: 2
    dim_hidden: 128
    n_layers: 2
    max_seq_len: 400
    """
    model = instantiate(OmegaConf.create(params))['seq_encoder']

    x = get_data()
    out = model(x)

    assert isinstance(out, torch.Tensor) and out.shape == torch.Size([4, 72])


def test_transformer_hydra_init_with_sequential():
    params = """
seq_encoder: 
    _target_: ptls.nn.TransformerSeqEncoder
    trx_encoder:
        _target_: torch.nn.Sequential
        _args_:
        -
            _target_: ptls.nn.TrxEncoder
            embeddings:
                mcc_code:
                    in: 200
                    out: 48
                tr_type:
                    in: 100
                    out: 24
            numeric_values:
                amount: identity
        -
            _target_: ptls.nn.PBLinear
            in_features: 73
            out_features: 24
    input_size: 24
    n_heads: 2
    dim_hidden: 128
    n_layers: 2
    max_seq_len: 400
    """
    model = instantiate(OmegaConf.create(params))['seq_encoder']

    x = get_data()
    out = model(x)

    assert isinstance(out, torch.Tensor) and out.shape == torch.Size([4, 24])



import warnings

import pytorch_lightning as pl
from ptls.data_load.iterable_processing.seq_len_filter import SeqLenFilter
from ptls.data_load.iterable_processing.feature_filter import FeatureFilter
from ptls.data_load.iterable_processing.target_move import TargetMove
from ptls.data_load.iterable_processing.to_torch_tensor import ToTorch
from ptls.data_load import IterableChain
from torch.utils.data import DataLoader
from ptls.data_load import padded_collate
from typing import List, Dict
from sklearn.model_selection import train_test_split


class SeqToTargetDatamodule(pl.LightningDataModule):
    r"""pytorch-lightning data module for supervised training
    Parameters
    ----------
     dataset: List[Dict]
        dataset
     pl_module: pl.LightningModule
        Don't required
     min_seq_len: int. Default: 0.
        The minimal length of sequences used for training. The shorter sequences would be skipped.
     valid_size: float. Default: 0.05.
        Should be between 0.0 and 1.0 and represent the proportion of the dataset to include in the test split.
     train_num_workers: int. Default: 0.
        The number of workers for train dataloader. 0 = single-process loader
     train_batch_size: int. Default: 256.
        The number of samples (before splitting to subsequences) in each batch during training.
     valid_num_workers: int. Default: 0.
        The number of workers for validation dataloader. 0 = single-process loader
     valid_batch_size: int. Default: 256.
        The number of samples (before splitting to subsequences) in each batch during validation.
     target_col: str. Default: 'target'.
        The name of target column.
     random_state : int. Default : 42.
        Controls the shuffling applied to the data before applying the split.
     """
    def __init__(self,
                 dataset: List[dict],
                 pl_module: pl.LightningModule = None,
                 min_seq_len: int = 0,
                 valid_size: float = 0.05,
                 train_num_workers: int = 0,
                 train_batch_size: int = 256,
                 valid_num_workers: int = 0,
                 valid_batch_size: int = 256,
                 target_col: str = 'target',
                 random_state: int = 42):
        warnings.warn('Use `ptls.frames.PtlsDataModule` with `ptls.frames.supervised.SeqToTargetDataset` '
                      'or `ptls.frames.supervised.SeqToTargetIterableDataset`',
                      DeprecationWarning)

        super().__init__()
        self.dataset_train, self.dataset_valid = train_test_split(dataset,
                                                                  test_size=valid_size,
                                                                  random_state=random_state)
        self.min_seq_len=min_seq_len
        self.train_num_workers = train_num_workers
        self.train_batch_size = train_batch_size
        self.valid_num_workers = valid_num_workers
        self.valid_batch_size = valid_batch_size
        self.target_col = target_col
        self.post_proc = IterableChain(*self.build_iterable_processing())

    def prepare_data(self):
        self.dataset_train = list(self.post_proc(iter(self.dataset_train)))
        self.dataset_valid = list(self.post_proc(iter(self.dataset_valid)))

    def build_iterable_processing(self):
        yield SeqLenFilter(min_seq_len=self.min_seq_len)
        yield ToTorch()
        yield TargetMove(self.target_col)
        yield FeatureFilter(drop_non_iterable=True)

    def train_dataloader(self):
        return DataLoader(
            dataset=self.dataset_train,
            collate_fn=padded_collate,
            num_workers=self.train_num_workers,
            batch_size=self.train_batch_size
        )

    def val_dataloader(self):
        return DataLoader(
            dataset=self.dataset_valid,
            collate_fn=padded_collate,
            num_workers=self.valid_num_workers,
            batch_size=self.valid_batch_size
        )

    def get_test_dataloader(self, data, num_workers=0, batch_size=128):
         return DataLoader(
             dataset=list(self.post_proc(iter(data))),
             collate_fn=padded_collate,
             num_workers=num_workers,
             batch_size=batch_size,
         )

